"""Reasoner API models."""
from typing import Optional, Set

from pydantic import constr, Field

from .base_model import BaseModel
from .utils import HashableSequence, HashableSet
from .results import Result
from .qgraph import QueryGraph
from .kgraph import KnowledgeGraph
from .shared import LogEntry, LogLevel
from .workflow import Workflow


class Message(BaseModel):
    """Message."""

    query_graph: Optional[QueryGraph] = Field(
        None,
        title="query graph",
        nullable=True,
    )
    knowledge_graph: Optional[KnowledgeGraph] = Field(
        None,
        title="knowledge graph",
        nullable=True,
    )
    results: Optional[HashableSet[Result]] = Field(
        None,
        title="list of results",
        nullable=True,
    )

    class Config:
        title = "message"
        extra = "forbid"

    def update(self, other):
        if hash(self.query_graph) != hash(other.query_graph):
            raise NotImplementedError("Query graph merging not supported yet")
        if other.knowledge_graph:
            if not self.knowledge_graph:
                self.knowledge_graph = KnowledgeGraph(nodes=[], edges=[])
            self.knowledge_graph.update(other.knowledge_graph)
        if other.results:
            self.results.update(other.results)


class Query(BaseModel):
    """Request."""

    message: Message = Field(
        ...,
        title="message",
    )
    log_level: Optional[LogLevel] = Field(
        None,
        title="log_level",
        nullable=True,
    )
    workflow: Optional[Workflow]

    class Config:
        title = "query"
        extra = "allow"
        schema_extra = {"x-body-name": "request_body"}


class AsyncQuery(BaseModel):
    """AsyncQuery."""

    callback: constr(regex=r"^https?://") = Field(..., format="uri")
    message: Message = Field(
        ...,
        title="message",
    )
    log_level: Optional[LogLevel] = Field(
        None,
        title="log_level",
        nullable=True,
    )
    workflow: Optional[Workflow]

    class Config:
        title = "query"
        extra = "allow"
        schema_extra = {"x-body-name": "request_body"}


class Response(BaseModel):
    """Response."""

    message: Message = Field(
        ...,
        title="message",
    )

    logs: Optional[HashableSequence[LogEntry]] = Field(None, nullable=True)

    status: Optional[str] = Field(None, nullable=True)

    workflow: Optional[Workflow]

    class Config:
        title = "response"
        extra = "allow"



"""Results models."""
from typing import Optional

from pydantic import Field

from .base_model import BaseModel
from .utils import HashableMapping, HashableSequence, HashableSet
from .shared import Attribute, CURIE


class EdgeBinding(BaseModel):
    """Edge binding."""

    id: str = Field(
        ...,
        title="knowledge graph id",
    )

    attributes: Optional[HashableSet[Attribute]] = Field(None, nullable=True)

    class Config:
        title = "edge binding"
        schema_extra = {
            "example": {
                "id": "string",
            },
        }
        extra = "allow"


class NodeBinding(BaseModel):
    """Node binding."""

    id: CURIE = Field(
        ...,
        title="knowledge graph id",
    )

    class Config:
        title = "node binding"
        schema_extra = {
            "example": {
                "id": "x:string",
            },
        }
        extra = "allow"


class Result(BaseModel):
    """Result."""

    node_bindings: HashableMapping[str, HashableSet[NodeBinding]] = Field(
        ...,
        title="list of node bindings",
    )
    edge_bindings: HashableMapping[str, HashableSet[EdgeBinding]] = Field(
        ...,
        title="list of edge bindings",
    )
    score: Optional[float] = Field(
        None,
        format="float",
    )

    class Config:
        title = "result"
        extra = "allow"



"""Knowledge graph models."""
from typing import Optional, Union

from pydantic import Field

from .shared import Attribute, BiolinkEntity, BiolinkPredicate, CURIE, EdgeIdentifier
from .base_model import BaseModel
from .utils import HashableMapping, HashableSequence, HashableSet


class Node(BaseModel):
    """Knowledge graph node."""

    categories: Optional[HashableSet[BiolinkEntity]] = Field(
        None,
        title="categories",
        nullable=True,
    )
    name: Optional[str] = Field(None, nullable=True)
    attributes: Optional[HashableSet[Attribute]] = Field(None, nullable=True)

    class Config:
        title = "knowledge-graph node"
        schema_extra = {
            "example": {
                "category": "string",
            },
        }
        extra = "forbid"
        allow_population_by_field_name = True

    def update(self, other):
        if other.name:
            self.name = other.name

        if other.categories:
            if self.categories:
                self.categories.update(other.categories)
            else:
                self.categories = other.categories

        if other.attributes:
            if self.attributes:
                self.attributes.update(other.attributes)
            else:
                self.attributes = other.attributes


class Edge(BaseModel):
    """Knowledge graph edge."""

    subject: CURIE = Field(
        ...,
        title="subject node id",
    )
    object: CURIE = Field(
        ...,
        title="object node id",
    )
    predicate: Optional[BiolinkPredicate] = Field(None, nullable=True)
    attributes: Optional[HashableSet[Attribute]] = Field(None, nullable=True)

    class Config:
        title = "knowledge-graph edge"
        extra = "forbid"

    def update(self, other):
        if other.attributes:
            if self.attributes:
                self.attributes.update(other.attributes)
            else:
                self.attributes = other.attributes


class KnowledgeGraph(BaseModel):
    """Knowledge graph."""

    nodes: HashableMapping[CURIE, Node] = Field(
        ...,
        title="nodes",
    )
    edges: HashableMapping[EdgeIdentifier, Edge] = Field(
        ...,
        title="edges",
    )

    class Config:
        title = "knowledge graph"
        extra = "allow"

    def update(self, other):
        for key, value in other.nodes.items():
            existing = self.nodes.get(key, None)
            if existing:
                existing.update(value)
            else:
                self.nodes[key] = value

        for key, value in other.edges.items():
            existing = self.edges.get(key, None)
            if existing:
                existing.update(value)
            else:
                self.edges[key] = value



from pydantic import validator
from reasoner_pydantic.shared import CURIE
from typing import Optional

from reasoner_pydantic import BiolinkEntity, BiolinkPredicate

from .base_model import BaseModel
from .utils import HashableMapping, HashableSequence, nonzero_validator


class MetaAttribute(BaseModel):
    """MetaAttribute."""

    attribute_type_id: CURIE
    attribute_source: Optional[str]
    original_attribute_names: Optional[HashableSequence[str]]
    constraint_use: Optional[bool] = False
    constraint_name: Optional[str]


class MetaNode(BaseModel):
    id_prefixes: HashableSequence[str]
    _nonzero_id_prefixes = validator("id_prefixes", allow_reuse=True)(nonzero_validator)
    attributes: Optional[HashableSequence[MetaAttribute]]

    class Config:
        extra = "forbid"


class MetaEdge(BaseModel):
    subject: BiolinkEntity
    predicate: BiolinkPredicate
    object: BiolinkEntity
    attributes: Optional[HashableSequence[MetaAttribute]]

    class Config:
        extra = "forbid"


class MetaKnowledgeGraph(BaseModel):
    nodes: HashableMapping[str, MetaNode]
    edges: HashableSequence[MetaEdge]



"""Shared models."""
from datetime import datetime
from enum import Enum
from typing import Any, Optional

from pydantic import constr, Field
from pydantic.class_validators import validator

from .base_model import BaseModel
from .utils import HashableSequence, make_hashable


class CURIE(str):
    """Compact URI."""


class EdgeIdentifier(str):
    """Identifier for an edge in a knowledge graph"""


class SubAttribute(BaseModel):
    """Attribute subattribute."""

    attribute_type_id: CURIE = Field(..., title="type")
    value: Any = Field(..., title="value")
    _make_value_hashable = validator("value", allow_reuse=True)(make_hashable)
    value_type_id: Optional[CURIE] = Field(
        None,
        title="value_type_id",
        nullable=True,
    )
    original_attribute_name: Optional[str] = Field(None, nullable=True)
    value_url: Optional[str] = Field(None, nullable=True)
    attribute_source: Optional[str] = Field(None, nullable=True)
    description: Optional[str] = Field(None, nullable=True)

    class Config:
        extra = "forbid"


class Attribute(BaseModel):
    """Node/edge attribute."""

    attribute_type_id: CURIE = Field(..., title="type")
    value: Any = Field(..., title="value")
    _make_value_hashable = validator("value", allow_reuse=True)(make_hashable)
    value_type_id: Optional[CURIE] = Field(
        None,
        title="value_type_id",
        nullable=True,
    )
    original_attribute_name: Optional[str] = Field(None, nullable=True)
    value_url: Optional[str] = Field(None, nullable=True)
    attribute_source: Optional[str] = Field(None, nullable=True)
    description: Optional[str] = Field(None, nullable=True)
    attributes: Optional[HashableSequence[SubAttribute]] = Field(None, nullable=True)

    class Config:
        extra = "forbid"


class BiolinkEntity(BaseModel):
    """Biolink entity."""

    __root__: constr(regex="^biolink:[A-Z][a-zA-Z]*$")

    class Config:
        title = "biolink entity"

    def __hash__(self):
        """Optimized hash function"""
        return hash(self.__root__)


class BiolinkPredicate(BaseModel):
    """Biolink predicate."""

    __root__: constr(regex="^biolink:[a-z][a-z_]*$")

    class Config:
        title = "biolink predicate"

    def __hash__(self):
        """Optimized hash function"""
        return hash(self.__root__)


class LogLevelEnum(str, Enum):
    """Log level."""

    error = "ERROR"
    warning = "WARNING"
    info = "INFO"
    debug = "DEBUG"


class LogLevel(BaseModel):
    """Log level."""

    __root__: LogLevelEnum


class LogEntry(BaseModel):
    """Log entry."""

    timestamp: Optional[datetime] = Field(None, nullable=True)
    level: Optional[LogLevel] = Field(None, nullable=True)
    code: Optional[str] = Field(None, nullable=True)
    message: Optional[str] = Field(None, nullable=True)

    class Config:
        extra = "allow"



"""Operations models."""
from enum import Enum
from typing import Any, Optional, Union
from pydantic.class_validators import validator

from pydantic.types import confloat, conint

from .base_model import BaseModel
from .utils import HashableMapping, HashableSequence, nonzero_validator


def constant(s: str):
    """Generate a static enum."""
    return Enum(value=s, names={s: s}, type=str)


class OperationAnnotate(BaseModel):
    id: constant("annotate")
    parameters: Optional[Any]

    class Config:
        extra = "forbid"


class AnnotateEdgesParameters(BaseModel):
    attributes: Optional[HashableSequence[str]]


class OperationAnnotateEdges(BaseModel):
    id: constant("annotate_edges")
    parameters: Optional[AnnotateEdgesParameters]

    class Config:
        extra = "forbid"


class AnnotateNodesParameters(BaseModel):
    attributes: Optional[HashableSequence[str]]


class OperationAnnotateNodes(BaseModel):
    id: constant("annotate_nodes")
    parameters: Optional[AnnotateNodesParameters]

    class Config:
        extra = "forbid"


class OperationBind(BaseModel):
    id: constant("bind")
    parameters: Optional[Any]

    class Config:
        extra = "forbid"


class OperationCompleteResults(BaseModel):
    id: constant("complete_results")
    parameters: Optional[Any]

    class Config:
        extra = "forbid"


class EnrichResultsParameters(BaseModel):
    pvalue_threshold: confloat(ge=0.0, le=1.0) = 1e-6
    qnode_keys: Optional[HashableSequence[str]]


class OperationEnrichResults(BaseModel):
    id: constant("enrich_results")
    parameters: Optional[EnrichResultsParameters]

    class Config:
        extra = "forbid"


class FillAllowParameters(BaseModel):
    allowlist: HashableSequence[str]
    _nonzero_allowlist = validator("allowlist", allow_reuse=True)(nonzero_validator)

    class Config:
        extra = "forbid"


class FillDenyParameters(BaseModel):
    denylist: HashableSequence[str]
    _nonzero_denylist = validator("denylist", allow_reuse=True)(nonzero_validator)

    class Config:
        extra = "forbid"


class FillParameters(BaseModel):
    _root_: Union[FillAllowParameters, FillDenyParameters]


class OperationFill(BaseModel):
    id: constant("fill")
    parameters: Optional[FillParameters]

    class Config:
        extra = "forbid"


class OperationFilterResults(BaseModel):
    id: constant("filter_results")
    parameters: Optional[Any]

    class Config:
        extra = "forbid"


class FilterResultsTopNParameters(BaseModel):
    max_results: conint(ge=0)

    class Config:
        extra = "forbid"


class OperationFilterResultsTopN(BaseModel):
    id: constant("filter_results_top_n")
    parameters: Optional[FilterResultsTopNParameters]

    class Config:
        extra = "forbid"


class OperationFilterKgraph(BaseModel):
    id: constant("filter_kgraph")
    parameters: Optional[Any]

    class Config:
        extra = "forbid"


class AboveOrBelowEnum(str, Enum):
    """'above' or 'below'."""

    above = "above"
    below = "below"


class FilterKgraphContinuousKedgeAttributeParameters(BaseModel):
    edge_attribute: str
    threshold: float
    remove_above_or_below: AboveOrBelowEnum
    qedge_keys: Optional[HashableSequence[str]]
    qnode_keys: HashableSequence[str] = []


class OperationFilterKgraphContinuousKedgeAttribute(BaseModel):
    id: constant("filter_kgraph_continuous_kedge_attribute")
    parameters: FilterKgraphContinuousKedgeAttributeParameters

    class Config:
        extra = "forbid"


class FilterKgraphDiscreteKedgeAttributeParameters(BaseModel):
    edge_attribute: str
    remove_value: Any
    qedge_keys: Optional[HashableSequence[str]]
    qnode_keys: HashableSequence[str] = []


class OperationFilterKgraphDiscreteKedgeAttribute(BaseModel):
    id: constant("filter_kgraph_discrete_kedge_attribute")
    parameters: FilterKgraphDiscreteKedgeAttributeParameters

    class Config:
        extra = "forbid"


class FilterKgraphDiscreteKnodeAttributeParameters(BaseModel):
    node_attribute: str
    remove_value: Any
    qnode_keys: Optional[HashableSequence[str]]


class OperationFilterKgraphDiscreteKnodeAttribute(BaseModel):
    id: constant("filter_kgraph_discrete_knode_attribute")
    parameters: FilterKgraphDiscreteKnodeAttributeParameters

    class Config:
        extra = "forbid"


class OperationFilterKgraphOrphans(BaseModel):
    id: constant("filter_kgraph_orphans")
    parameters: Optional[Any]

    class Config:
        extra = "forbid"


class TopOrBottomEnum(str, Enum):
    """'top' or 'bottom'."""

    top = "top"
    bottom = "bottom"


class FilterKgraphTopNParameters(BaseModel):
    edge_attribute: str
    max_edges: conint(le=0) = 50
    remove_top_or_bottom: TopOrBottomEnum = TopOrBottomEnum.top
    qedge_keys: Optional[HashableSequence[str]]
    qnode_keys: HashableSequence[str] = []


class FilterKgraphPercentileParameters(BaseModel):
    edge_attribute: str
    threshold: confloat(ge=0, le=100) = 95
    remove_above_or_below: AboveOrBelowEnum = AboveOrBelowEnum.below
    qedge_keys: Optional[HashableSequence[str]]
    qnode_keys: HashableSequence[str] = []


class OperationFilterKgraphPercentile(BaseModel):
    id: constant("filter_kgraph_percentile")
    parameters: FilterKgraphPercentileParameters

    class Config:
        extra = "forbid"


class PlusOrMinusEnum(str, Enum):
    """'plus' or 'minus'."""

    plus = "plus"
    minus = "minus"


class FilterKgraphStdDevParameters(BaseModel):
    edge_attribute: str
    plus_or_minus_std_dev: PlusOrMinusEnum = PlusOrMinusEnum.plus
    num_sigma: confloat(ge=0) = 1
    remove_above_or_below: AboveOrBelowEnum = AboveOrBelowEnum.below
    qedge_keys: Optional[HashableSequence[str]]
    qnode_keys: HashableSequence[str] = []


class OperationFilterKgraphStdDev(BaseModel):
    id: constant("filter_kgraph_std_dev")
    parameters: FilterKgraphStdDevParameters

    class Config:
        extra = "forbid"


class OperationFilterKgraphTopN(BaseModel):
    id: constant("filter_kgraph_top_n")
    parameters: FilterKgraphTopNParameters

    class Config:
        extra = "forbid"


class OperationLookup(BaseModel):
    id: constant("lookup")
    parameters: Optional[Any]

    class Config:
        extra = "forbid"


class OperationOverlay(BaseModel):
    id: constant("overlay")
    parameters: Optional[Any]

    class Config:
        extra = "forbid"


class OverlayComputeJaccardParameters(BaseModel):
    intermediate_node_key: str
    end_node_keys: HashableSequence[str]
    virtual_relation_label: str


class OperationOverlayComputeJaccard(BaseModel):
    id: constant("overlay_compute_jaccard")
    parameters: OverlayComputeJaccardParameters

    class Config:
        extra = "forbid"


class OverlayComputeNgdParameters(BaseModel):
    qnode_keys: HashableSequence[str]
    virtual_relation_label: str


class OperationOverlayComputeNgd(BaseModel):
    id: constant("overlay_compute_ngd")
    parameters: OverlayComputeNgdParameters

    class Config:
        extra = "forbid"


class OperationOverlayConnectKnodes(BaseModel):
    id: constant("overlay_connect_knodes")
    parameters: Optional[Any]

    class Config:
        extra = "forbid"


class OverlayFisherExactTestParameters(BaseModel):
    subject_qnode_key: str
    object_qnode_key: str
    virtual_relation_label: str
    rel_edge_key: Optional[str]


class OperationOverlayFisherExactTest(BaseModel):
    id: constant("overlay_fisher_exact_test")
    parameters: OverlayFisherExactTestParameters

    class Config:
        extra = "forbid"


class OperationRestate(BaseModel):
    id: constant("restate")
    parameters: Optional[Any]

    class Config:
        extra = "forbid"


class OperationScore(BaseModel):
    id: constant("score")
    parameters: Optional[Any]

    class Config:
        extra = "forbid"


class OperationSortResults(BaseModel):
    id: constant("sort_results")
    parameters: Optional[Any]

    class Config:
        extra = "forbid"


class AscOrDescEnum(str, Enum):
    """'ascending' or 'descending'."""

    ascending = "ascending"
    descending = "descending"


class SortResultsEdgeAttributeParameters(BaseModel):
    edge_attribute: str
    ascending_or_descending: AscOrDescEnum
    qedge_keys: Optional[HashableSequence[str]]


class OperationSortResultsEdgeAttribute(BaseModel):
    id: constant("sort_results_edge_attribute")
    parameters: SortResultsEdgeAttributeParameters

    class Config:
        extra = "forbid"


class SortResultsNodeAttributeParameters(BaseModel):
    node_attribute: str
    ascending_or_descending: AscOrDescEnum
    qnode_keys: Optional[HashableSequence[str]]


class OperationSortResultsNodeAttribute(BaseModel):
    id: constant("sort_results_node_attribute")
    parameters: SortResultsNodeAttributeParameters

    class Config:
        extra = "forbid"


class SortResultsScoreParameters(BaseModel):
    ascending_or_descending: AscOrDescEnum


class OperationSortResultsScore(BaseModel):
    id: constant("sort_results_score")
    parameters: SortResultsScoreParameters

    class Config:
        extra = "forbid"


operations = [
    OperationAnnotate,
    OperationAnnotateEdges,
    OperationAnnotateNodes,
    OperationBind,
    OperationCompleteResults,
    OperationEnrichResults,
    OperationFill,
    OperationFilterKgraph,
    OperationFilterKgraphContinuousKedgeAttribute,
    OperationFilterKgraphDiscreteKedgeAttribute,
    OperationFilterKgraphDiscreteKnodeAttribute,
    OperationFilterKgraphOrphans,
    OperationFilterKgraphPercentile,
    OperationFilterKgraphStdDev,
    OperationFilterKgraphTopN,
    OperationFilterResults,
    OperationFilterResultsTopN,
    OperationLookup,
    OperationOverlay,
    OperationOverlayComputeJaccard,
    OperationOverlayComputeNgd,
    OperationOverlayConnectKnodes,
    OperationOverlayFisherExactTest,
    OperationRestate,
    OperationScore,
    OperationSortResults,
    OperationSortResultsEdgeAttribute,
    OperationSortResultsNodeAttribute,
    OperationSortResultsScore,
]


class Operation(BaseModel):
    __root__: Union[
        OperationAnnotate,
        OperationAnnotateEdges,
        OperationAnnotateNodes,
        OperationBind,
        OperationCompleteResults,
        OperationEnrichResults,
        OperationFill,
        OperationFilterKgraph,
        OperationFilterKgraphContinuousKedgeAttribute,
        OperationFilterKgraphDiscreteKedgeAttribute,
        OperationFilterKgraphDiscreteKnodeAttribute,
        OperationFilterKgraphOrphans,
        OperationFilterKgraphPercentile,
        OperationFilterKgraphStdDev,
        OperationFilterKgraphTopN,
        OperationFilterResults,
        OperationFilterResultsTopN,
        OperationLookup,
        OperationOverlay,
        OperationOverlayComputeJaccard,
        OperationOverlayComputeNgd,
        OperationOverlayConnectKnodes,
        OperationOverlayFisherExactTest,
        OperationRestate,
        OperationScore,
        OperationSortResults,
        OperationSortResultsEdgeAttribute,
        OperationSortResultsNodeAttribute,
        OperationSortResultsScore,
    ]


class Workflow(BaseModel):
    __root__: HashableSequence[Operation]



"""Setup file for reasoner package."""
from setuptools import setup

setup(
    name="reasoner-pydantic",
    version="2.0.0",
    author="Patrick Wang",
    author_email="patrick@covar.com",
    url="https://github.com/TranslatorSRI/reasoner-pydantic",
    description="Pydantic models for the Reasoner API data formats",
    packages=["reasoner_pydantic"],
    include_package_data=True,
    install_requires=["pydantic>=1.8"],
    zip_safe=False,
    license="MIT",
    python_requires=">=3.6",
)



import collections
from typing import Callable, Dict, List, Generic, Set, TypeVar

from pydantic import PrivateAttr
from pydantic.generics import GenericModel

KeyType = TypeVar("KeyType")
ValueType = TypeVar("ValueType")


class HashableMapping(
    GenericModel,
    Generic[KeyType, ValueType],
    collections.abc.MutableMapping,
):
    """
    Custom class that implements MutableMapping and is hashable

    Hash will be recomputed if items are updated or deleted. The
    hash can be considered valid if the values are immutable.
    """

    __root__: Dict[KeyType, ValueType]

    _hash: int = PrivateAttr(default=None)
    _invalidate_hook: Callable = PrivateAttr(default=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for value in self.__root__.values():
            if hasattr(value, "_invalidate_hook"):
                value._invalidate_hook = self.invalidate_hash

    def __getitem__(self, k):
        return self.__root__[k]

    def __iter__(self):
        return iter(self.__root__)

    def __len__(self):
        return len(self.__root__)

    def __setitem__(self, k, v):
        self.invalidate_hash()
        if hasattr(v, "_invalidate_hook"):
            v._invalidate_hook = self.invalidate_hash
        self.__root__[k] = v

    def __delitem__(self, k):
        self.invalidate_hash()
        del self.__root__[k]

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(tuple((k, v) for k, v in self.__root__.items()))
        return self._hash

    def __eq__(self, other) -> bool:
        return self.__hash__() == other.__hash__()

    def invalidate_hash(self):
        """Invalidate stored hash value"""
        self._hash = None
        # Propogate
        if self._invalidate_hook:
            self._invalidate_hook()


class HashableSequence(
    GenericModel,
    Generic[ValueType],
    collections.abc.MutableSequence,
):
    """
    Custom class that implements MutableSequence and is hashable

    Hash will be recomputed if items are updated or deleted. The
    hash can be considered valid if the values are immutable.
    """

    __root__: List[ValueType]
    _hash: int = PrivateAttr(default=None)
    _invalidate_hook: Callable = PrivateAttr(default=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for value in self.__root__:
            if hasattr(value, "_invalidate_hook"):
                value._invalidate_hook = self.invalidate_hash

    def __getitem__(self, i):
        return self.__root__[i]

    def __iter__(self):
        return iter(self.__root__)

    def __len__(self):
        return len(self.__root__)

    def __setitem__(self, i, v):
        self.invalidate_hash()
        if hasattr(v, "_invalidate_hook"):
            v._invalidate_hook = self.invalidate_hash
        self.__root__[i] = v

    def __delitem__(self, i):
        self.invalidate_hash()
        del self.__root__[i]

    def insert(self, i, v):
        self.invalidate_hash()
        self.__root__.insert(i, v)

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(tuple(self.__root__))
        return self._hash

    def __eq__(self, other) -> bool:
        return self.__hash__() == other.__hash__()

    def invalidate_hash(self):
        """Invalidate stored hash value"""
        self._hash = None
        # Propogate
        if self._invalidate_hook:
            self._invalidate_hook()


class HashableSet(
    GenericModel,
    Generic[ValueType],
    collections.abc.MutableSet,
):
    """
    Custom class that implements MutableSet and is hashable

    Hash will be recomputed if items are updated or deleted. The
    hash can be considered valid if the values are immutable.
    """

    __root__: Set[ValueType]
    _hash: int = PrivateAttr(default=None)
    _invalidate_hook: Callable = PrivateAttr(default=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for value in self.__root__:
            if hasattr(value, "_invalidate_hook"):
                value._invalidate_hook = self.invalidate_hash

    def __contains__(self, v):
        return v in self.__root__

    def __iter__(self):
        return iter(self.__root__)

    def __len__(self):
        return len(self.__root__)

    def add(self, v):
        self.invalidate_hash()
        if hasattr(v, "_invalidate_hook"):
            v._invalidate_hook = self.invalidate_hash
        self.__root__.add(v)

    def update(self, other):
        self.invalidate_hash()
        self.__root__.update(other)

    def discard(self, v):
        self.invalidate_hash()
        self.__root__.discard(v)

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(tuple(self.__root__))
        return self._hash

    def invalidate_hash(self):
        """Invalidate stored hash value"""
        self._hash = None
        # Propogate
        if self._invalidate_hook:
            self._invalidate_hook()

    def dict(self, *args, **kwargs):
        """Custom serialization method to convert to list"""

        # Normally, the dict method tries to cast to the __root__ type.
        # This isn't an issue for most __root__ types, but here that causes:
        # set({"hello" : "world"}) which doesn't work because dicts are not hashable
        # This overrides that functionality to cast to a list instead
        return [self._get_value(v, to_dict=True, **kwargs) for v in self]


def nonzero_validator(v):
    if v != None and len(v) == 0:
        raise ValueError("Must have nonzero number of elements")
    return v


def make_hashable(o):
    """
    Convert a generic Python object to a hashable one recursively

    This is an expensive operation, so it is best used sparingly
    """

    # type(o) is faster than isinstance(o) because it doesn't
    # traverse the inheritance hierarchy
    o_type = str(type(o))

    if "dict" in o_type:
        return HashableMapping.parse_obj(((k, make_hashable(v)) for k, v in o.items()))
    if "list" in o_type:
        return HashableSequence.parse_obj(make_hashable(v) for v in o)

    return o



from typing import Callable
import weakref
from pydantic import BaseModel as PydanticBaseModel, PrivateAttr


class BaseModel(PydanticBaseModel):
    """
    Custom base model for all classes

    This provides a hash function that assumes all fields are either:

    1. Immutable
    2. Derived from BaseModel
    3. Able to call a hash invalidation hook on their own
    """

    _hash: int = PrivateAttr(default=None)
    _invalidate_hook: Callable = PrivateAttr(default=None)

    def __hash__(self) -> int:
        """Hash function based on Pydantic implementation"""
        if not self._hash:
            self._hash = hash((self.__class__, tuple(self.__dict__.values())))
        return self._hash

    def __eq__(self, other) -> bool:
        """Equality function that calls hash function"""
        return self.__hash__() == other.__hash__()

    def __setattr__(self, name, value):
        """Custom setattr that invalidates hash"""

        if name != "_hash":
            self.invalidate_hash()
        return super().__setattr__(name, value)

    def invalidate_hash(self):
        """Invalidate stored hash value"""
        self._hash = None
        # Propogate
        if self._invalidate_hook:
            self._invalidate_hook()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.__custom_root_type__:
            return

        # Look for BaseModel fields and give them a hook
        # that they can use to invalidate hash on this object
        for value in self.__dict__.values():
            if hasattr(value, "_invalidate_hook"):
                value._invalidate_hook = self.invalidate_hash

    def update(self, other):
        """Update fields on this object with fields from other object"""
        raise NotImplementedError(
            f"Model {self.__class__.__name__} has no update method"
        )

    class Config:
        validate_assignment = True



"""Query graph models."""
from enum import Enum

from pydantic.class_validators import validator
from reasoner_pydantic.utils import HashableMapping
from typing import Any, Optional, Union

from pydantic import Field

from .base_model import BaseModel
from .utils import HashableMapping, HashableSequence, nonzero_validator
from .shared import BiolinkEntity, BiolinkPredicate, CURIE


class Operator(str, Enum):
    """Operator."""

    equal_to = "=="
    greater_than = ">"
    less_than = "<"
    matches = "matches"


class QueryConstraint(BaseModel):
    """QNode or QEdge constraint."""

    name: str = Field(
        ...,
        title="name",
        nullable=False,
    )
    id: CURIE = Field(
        ...,
        title="id",
        nullable=False,
    )
    negated: bool = Field(
        False,
        title="not",
        alias="not",
    )
    operator: Operator = Field(
        ...,
        title="operator",
    )
    value: Any = Field(
        ...,
        title="value",
    )
    unit_id: Optional[Any] = Field(
        None,
        title="unit_id",
    )
    unit_name: Optional[Any] = Field(
        None,
        title="unit_name",
    )

    class Config:
        extra = "forbid"

    def dict(self, *args, **kwargs):
        output = super().dict(*args, **kwargs)
        output["not"] = output.pop("negated", False)
        return output


class QNode(BaseModel):
    """Query node."""

    ids: Optional[HashableSequence[CURIE]] = Field(
        None,
        title="ids",
        nullable=True,
    )
    _nonzero_ids = validator("ids", allow_reuse=True)(nonzero_validator)

    categories: Optional[HashableSequence[BiolinkEntity]] = Field(
        None,
        title="categories",
        nullable=True,
    )
    _nonzero_categories = validator("categories", allow_reuse=True)(nonzero_validator)

    is_set: bool = False
    constraints: Optional[HashableSequence[QueryConstraint]] = Field(
        HashableSequence.parse_obj([]),
        title="constraints",
    )

    class Config:
        title = "query-graph node"
        extra = "allow"
        allow_population_by_field_name = True


class QEdge(BaseModel):
    """Query edge."""

    subject: str = Field(
        ...,
        title="subject node id",
    )
    object: str = Field(
        ...,
        title="object node id",
    )

    predicates: Union[HashableSequence[BiolinkPredicate], None] = Field(
        None,
        title="predicates",
        nullable=True,
    )
    _nonzero_predicates = validator("predicates", allow_reuse=True)(nonzero_validator)

    constraints: Optional[HashableSequence[QueryConstraint]] = Field(
        title="constraints",
    )

    class Config:
        title = "query-graph edge"
        extra = "allow"
        allow_population_by_field_name = True


class QueryGraph(BaseModel):
    """Query graph."""

    nodes: HashableMapping[str, QNode] = Field(
        ...,
        title="dict of nodes",
    )
    edges: HashableMapping[str, QEdge] = Field(
        ...,
        title="dict of edges",
    )

    class Config:
        title = "simple query graph"
        extra = "allow"
